@@ -42,6 +42,7 @@ use digest::DynDigest;
 use hmac_sha256::Hash as Sha256;
 use hmac_sha512::sha384::Hash as Sha384;
 use hmac_sha512::Hash as Sha512;
+use num_padding::ToBytesPadded;
 use rand::Rng;
 use rsa::algorithms::mgf1_xor;
 use rsa::internals as rsa_internals;
@@ -56,6 +57,8 @@ pub mod reexports {
     pub use {digest, hmac_sha512, rand, rsa};
 }
 
+mod num_padding;
+
 #[derive(Debug, Copy, Clone, Eq, PartialEq)]
 pub enum Error {
     InternalError,
@@ -63,6 +66,7 @@ pub enum Error {
     VerificationFailed,
     EncodingError,
     InvalidKey,
+    IncompatibleParameters,
 }
 
 impl std::error::Error for Error {}
@@ -75,6 +79,7 @@ impl Display for Error {
             Error::VerificationFailed => write!(f, "Verification failed"),
             Error::EncodingError => write!(f, "Encoding error"),
             Error::InvalidKey => write!(f, "Invalid key"),
+            Error::IncompatibleParameters => write!(f, "Incompatible parameters"),
         }
     }
 }
@@ -290,6 +295,146 @@ impl PublicKey {
             .into())
     }
 
+    const fn spki_tpl() -> &'static [u8] {
+        const SEQ: u8 = 0x30;
+        const EXT: u8 = 0x80;
+        const CON: u8 = 0xa0;
+        const INT: u8 = 0x02;
+        const BIT: u8 = 0x03;
+        const OBJ: u8 = 0x06;
+        const TPL: &[u8] = &[
+            SEQ,
+            EXT | 2,
+            0,
+            0, // container length - offset 2
+            SEQ,
+            61, // Algorithm sequence
+            OBJ,
+            9,
+            0x2a,
+            0x86,
+            0x48,
+            0x86,
+            0xf7,
+            0x0d,
+            0x01,
+            0x01,
+            0x0a, // Signature algorithm (RSASSA-PSS)
+            SEQ,
+            48, // RSASSA-PSS parameters sequence
+            CON | 0,
+            2 + 2 + 9,
+            SEQ,
+            2 + 9,
+            OBJ,
+            9,
+            0,
+            0,
+            0,
+            0,
+            0,
+            0,
+            0,
+            0,
+            0, // Hash function - offset 21
+            CON | 1,
+            2 + 24,
+            SEQ,
+            24,
+            OBJ,
+            9,
+            0x2a,
+            0x86,
+            0x48,
+            0x86,
+            0xf7,
+            0x0d,
+            0x01,
+            0x01,
+            0x08, // Padding function (MGF1) and parameters
+            SEQ,
+            2 + 9,
+            OBJ,
+            9,
+            0,
+            0,
+            0,
+            0,
+            0,
+            0,
+            0,
+            0,
+            0, // MGF1 hash function - offset 49
+            CON | 2,
+            2 + 1,
+            INT,
+            1,
+            0, // Salt length - offset 66
+            BIT,
+            EXT | 2,
+            0,
+            0, // Public key length - Bit string - offset 69
+            0, // No partial bytes
+        ];
+        TPL
+    }
+
+    pub fn to_spki(&self, options: Option<&Options>) -> Result<Vec<u8>, Error> {
+        let tpl = Self::spki_tpl();
+        let default_options = Options::default();
+        let options = options.unwrap_or(&default_options);
+        let der = self.to_der()?;
+        if der.len() <= 24 {
+            return Err(Error::EncodingError);
+        }
+        let raw = &der[24..];
+        let container_len = tpl.len() - 4 + raw.len();
+        let out_len = tpl.len() + raw.len();
+        let mut out = Vec::with_capacity(out_len);
+        out.extend_from_slice(tpl);
+        out.extend_from_slice(raw);
+        out[2..4].copy_from_slice(&(container_len as u16).to_be_bytes());
+        out[66] = options.salt_len() as u8;
+        out[69..71].copy_from_slice(&(1 + raw.len() as u16).to_be_bytes());
+        let mut mgf1_s: [u8; 13] = [48, 11, 6, 9, 96, 134, 72, 1, 101, 3, 4, 2, 0];
+        mgf1_s[12] = match options.hash {
+            Hash::Sha256 => 1,
+            Hash::Sha384 => 2,
+            Hash::Sha512 => 3,
+        };
+        out[21..][..mgf1_s.len()].copy_from_slice(&mgf1_s);
+        out[49..][..mgf1_s.len()].copy_from_slice(&mgf1_s);
+        Ok(out)
+    }
+
+    pub fn from_spki(spki: &[u8], _options: Option<&Options>) -> Result<Self, Error> {
+        if spki.len() > 800 {
+            return Err(Error::EncodingError);
+        }
+        let tpl = Self::spki_tpl();
+        if spki.len() <= tpl.len() {
+            return Err(Error::EncodingError);
+        }
+        if spki[6..18] != tpl[6..18] {
+            return Err(Error::EncodingError);
+        }
+        let alg_len = spki[5] as usize;
+        if spki.len() <= alg_len + 10 {
+            return Err(Error::EncodingError);
+        }
+        let raw = &spki[alg_len + 10..];
+        let der_seq: &mut [u8] = &mut [
+            0x30, 0x82, 0x01, 0x22, 0x30, 0x0d, 0x06, 0x09, 0x2a, 0x86, 0x48, 0x86, 0xf7, 0x0d,
+            0x01, 0x01, 0x01, 0x05, 0x00, 0x03, 0x82, 0x01, 0x0f,
+        ];
+        der_seq[2..][..2].copy_from_slice(&(raw.len() as u16 + 19).to_be_bytes());
+        der_seq[21..][..2].copy_from_slice(&(raw.len() as u16).to_be_bytes());
+        let mut der = Vec::with_capacity(der_seq.len() + raw.len());
+        der.extend_from_slice(der_seq);
+        der.extend_from_slice(raw);
+        Self::from_der(&der)
+    }
+
     /// Blind a message to be signed
     pub fn blind(&self, msg: impl AsRef<[u8]>, options: &Options) -> Result<BlindingResult, Error> {
         let msg = msg.as_ref();
@@ -320,8 +465,8 @@ impl PublicKey {
 
         let (blind_msg, secret) = rsa_internals::blind(&mut rng, self.as_ref(), &m);
         Ok(BlindingResult {
-            blind_msg: BlindedMessage(blind_msg.to_bytes_be()),
-            secret: Secret(secret.to_bytes_be()),
+            blind_msg: BlindedMessage(blind_msg.to_bytes_be_padded(modulus_bytes)),
+            secret: Secret(secret.to_bytes_be_padded(modulus_bytes)),
         })
     }
 
@@ -339,8 +484,10 @@ impl PublicKey {
         }
         let blind_sig = BigUint::from_bytes_be(blind_sig);
         let secret = BigUint::from_bytes_be(secret);
-        let sig =
-            Signature(rsa_internals::unblind(self.as_ref(), &blind_sig, &secret).to_bytes_be());
+        let sig = Signature(
+            rsa_internals::unblind(self.as_ref(), &blind_sig, &secret)
+                .to_bytes_be_padded(modulus_bytes),
+        );
         self.verify(&sig, msg, options)?;
         Ok(sig)
     }
@@ -428,8 +575,11 @@ impl SecretKey {
         }
         let mut rng = rand::thread_rng();
         let blind_msg = BigUint::from_bytes_be(blind_msg.as_ref());
+        if &blind_msg >= self.0.n() {
+            return Err(Error::UnsupportedParameters);
+        }
         let blind_sig = rsa_internals::decrypt_and_check(Some(&mut rng), self.as_ref(), &blind_msg)
             .map_err(|_| Error::InternalError)?;
-        Ok(BlindSignature(blind_sig.to_bytes_be()))
+        Ok(BlindSignature(blind_sig.to_bytes_be_padded(modulus_bytes)))
     }
 }
